@@ -18,12 +18,16 @@ from .teletypes import (
     Chat,
     ChatMember,
     BotCommand,
-    InlineQueryResult,
     WebhookInfo,
     InputMedia,
-    Poll)
+    Poll,
+    User
+    )
+
+from .teletypes.inline_query_result import InlineQueryResult
 
 from .commands import CommandContainer
+from .callbacks import Callbacks
 from .util import check_args
 
 from .api_request import (
@@ -79,9 +83,26 @@ from .api_request import (
     edit_message_media,
     edit_message_reply_markup,
     stop_poll,
-    delete_message)
+    delete_message,
+    send_media_group)
 
-from ..util import Listener
+from ..listener import Listener
+
+
+using_greenlets = False
+try:
+    import gevent
+    from gevent.monkey import is_module_patched
+    if any([
+            is_module_patched('socket'),
+            is_module_patched('threading'),
+            is_module_patched('time')
+            ]):
+        using_greenlets = True
+    else:
+        del gevent
+except ImportError:
+    pass
 
 
 class OrigamiBot:
@@ -90,7 +111,12 @@ class OrigamiBot:
         self.token = token
         self.webhook = None
         self.updates = deque()
+        self.inline = deque()
+        self.callback = deque()
         self.interval = 0.1
+        self.safe_poll = True
+        self._name = None
+        self.first_only_command = True
 
         self._listen_thread = StoppableThread(
             name='Listen thread',
@@ -100,41 +126,108 @@ class OrigamiBot:
         self._webhook_thread = StoppableThread(
             name='Webhook thread',
             target=self._webhook_loop,
-            daemon=True
-        )
+            daemon=True)
 
         self._process_thread = StoppableThread(
             name='Update process thread',
             target=self._process_updates_loop,
             daemon=True)
+        
+        self._inline_thread = StoppableThread(
+            name='Inline thread',
+            target=self._inline_loop,
+            daemon=True)
+        
+        self._callback_thread = StoppableThread(
+            name='Inline thread',
+            target=self._callback_loop,
+            daemon=True)
 
         self.has_updates = Event()
+        self.has_inline = Event()
+        self.has_callback = Event()
 
         self._last_update_id = 0
 
         self.command_container = CommandContainer()
+        self.inline_container = Callbacks()
+        self.callback_container = Callbacks()
         self.listeners = []
 
-    def start(self):
-        """Start listening for updates. Non-blocking!"""
-        if self.webhook is None:
-            self._listen_thread.start()
-        else:
+    @property
+    def name(self):
+        if self._name is None:
+            self._name = self.get_me().username
+        return self._name
+
+    def start(self,
+              webhook=False,
+              poll=True, 
+              process=True,
+              inline=True,
+              callback=True):
+        """Start listening for updates. Non-blocking!
+        
+        start_webhook=False - Start inbuilt webhook server
+        poll=True - Start Polling thread
+        process=True - Start update processing thread
+        inline=True - Start separate inline thread
+        """
+        if webhook and not poll:
             self._webhook_thread.start()
-        self._process_thread.start()
+        elif poll:
+            self._listen_thread.start()
+        if process:
+            self._process_thread.start()
+        if inline:
+            self._inline_thread.start()
+        if callback:
+            self._callback_thread.start()
 
     def stop(self):
         """Terminate all bot's threads."""
         if self._listen_thread.is_alive():
             self._listen_thread.stop()
             self._listen_thread.join()
+        if self._process_thread.is_alive():
             self._process_thread.stop()
             self._process_thread.join()
+        if self._inline_thread.is_alive():
+            self._inline_thread.stop()
+            self._inline_thread.join()
 
     def process_update(self, update: Update):
         """Process a single update."""
         if update.message is not None:
+            update.message.bot = self
+            self._call_listeners(
+                'on_message',
+                update.message)
             self._handle_message(update.message)
+        elif update.channel_post is not None:
+            self._call_listeners(
+                'on_channel_post',
+                update.channel_post)
+        elif update.edited_message is not None:
+            self._call_listeners(
+                'on_edited_message',
+                update.edited_message)
+        elif update.edited_channel_post is not None:
+            self._call_listeners(
+                'on_edited_channel_post',
+                update.edited_channel_post)
+        elif update.inline_query is not None:
+            if self._inline_thread.is_alive():
+                self.inline.append(update.inline_query)
+                self.has_inline.set()
+            else:
+                self.inline_container.call(update.inline_query)
+        elif update.callback_query is not None:
+            if self._callback_thread.is_alive():
+                self.callback.append(update.callback_query)
+                self.has_callback.set()
+            else:
+                self.callback_container.call(update.callback_query)
 
     def add_commands(self, obj):
         """Add an object to bot's commands container.
@@ -144,11 +237,26 @@ class OrigamiBot:
         """
         self.command_container.add_command(obj)
 
+    def remove_commands(self, obj):
+        self.command_container.remove(obj)
+
     def add_listener(self, obj):
         if not isinstance(obj, Listener):
             raise TypeError(f'{obj} is not a Listener')
         self.listeners.append(obj)
 
+    def add_inline(self, obj):
+        self.inline_container.add(obj)
+
+    def remove_inline(self, obj):
+        self.inline_container.remove(obj)
+
+    def add_callback(self, obj):
+        self.callback_container.add(obj)
+
+    def remove_callback(self, obj):
+        self.callback_container.remove(obj)
+
     def remove_commands_by_filter(self, filter_func: Callable):
         """Remove commands from container by filter
 
@@ -171,7 +279,7 @@ class OrigamiBot:
             self._last_update_id = updates[-1].update_id
         return updates
 
-    def get_me(self):
+    def get_me(self) -> User:
         return get_me(self.token)
 
     def send_message(self,
@@ -192,7 +300,27 @@ class OrigamiBot:
                             parse_mode,
                             disable_web_page_preview,
                             disable_notification,
-                            reply_to_message_id)
+                            reply_to_message_id,
+                            reply_markup)
+
+    def reply_to(self, 
+                 msg: Message,
+                 text: str, 
+                 parse_mode: Optional[str] = None,
+                 disable_web_page_preview: Optional[bool] = None,
+                 disable_notification: Optional[bool] = None,
+                 reply_markup: Optional[ReplyMarkup] = None) -> Message:
+        mid = msg.message_id
+        cid = msg.chat.id
+        return self.send_message(
+            cid,
+            text,
+            parse_mode=parse_mode,
+            disable_web_page_preview=disable_web_page_preview,
+            disable_notification=disable_notification,
+            reply_to_message_id=mid,
+            reply_markup=reply_markup
+        )
 
     def forward_message(self,
                         chat_id: Union[int, str],
@@ -622,6 +750,23 @@ class OrigamiBot:
             action
         )
 
+    def send_media_group(self,
+                         chat_id: Union[int, str],
+                         media: List[InputMedia],
+                         disable_notification: Optional[bool] = None,
+                         reply_to_message_id: Optional[bool] = None
+                         ) -> Message:
+        """Use this method to send a group of photos or videos as an album.
+        On success, an array of the sent Messages is returned
+        """
+        return send_media_group(
+            self.token,
+            chat_id,
+            media,
+            disable_notification,
+            reply_to_message_id
+        )
+
     def get_user_profile_photos(self,
                                 user_id: int,
                                 offset: Optional[int] = None,
@@ -1197,13 +1342,48 @@ class OrigamiBot:
             while True:
                 if current_thread().stopped:
                     break
-                updates = self.get_updates()
+                if self.safe_poll:
+                    try:
+                        updates = self.get_updates()
+                    except (IOError, ConnectionError) as err:
+                        self._call_listeners("on_poll_error", err)
+                        sleep(self.interval)
+                        continue
+                else:
+                    updates = self.get_updates()
+
                 if updates:
                     self.updates.extend(updates)
                     self.has_updates.set()
                 sleep(self.interval)
 
-    def _handle_commands(self, message: Message, first_only=False) -> bool:
+    def _inline_loop(self):
+        while True:
+            if current_thread().stopped:
+                break
+            self.has_inline.wait()
+            while self.inline:
+                query = self.inline.popleft()
+                if not using_greenlets:
+                    self.inline_container.call(query)
+                else:
+                    gevent.spawn(self.inline_container.call, query)
+            self.has_inline.clear()
+
+    def _callback_loop(self):
+        while True:
+            if current_thread().stopped:
+                break
+            self.has_callback.wait()
+            while self.callback:
+                query = self.callback.popleft()
+                if not using_greenlets:
+                    self.callback_container.call(query)
+                else:
+                    gevent.spawn(self.callback_container.call, query)
+            self.has_callback.clear()
+
+    def _handle_commands(self, message: Message) -> bool:
         """Check message for commands in it.
 
         If there are commands, returns True and tries execute them
@@ -1227,7 +1407,7 @@ class OrigamiBot:
             (text[e.offset:e.offset+e.length], e.offset, e.offset+e.length)
             for e in entities
             if e.type == 'bot_command'
-            and (e.offset == 0 or not first_only)
+            and (e.offset == 0 or not self.first_only_command)
             ])
 
         if not commands:
@@ -1235,6 +1415,10 @@ class OrigamiBot:
 
         while commands:
             command, start, end = commands.popleft()
+            if '@' in command:
+                command, bot_name = command.split('@', 1)
+                if bot_name != self.name:
+                    continue
 
             # Parse arguments for each command
             if commands:
@@ -1250,15 +1434,20 @@ class OrigamiBot:
                     self._call_listeners('on_command_failure', message)
                     continue
                 try:
-                    method(*bound_args.args, **bound_args.kwargs)
+                    if not using_greenlets:
+                        method(*bound_args.args, **bound_args.kwargs)
+                    else:
+                        gevent.spawn(
+                            method,
+                            *bound_args.args,
+                            **bound_args.kwargs
+                            )
                 except Exception as err:
                     self._call_listeners('on_command_failure', message, err)
         return True
 
     def _handle_message(self, message: Message):
         """Process a single message"""
-        self._call_listeners('on_message', message)
-
         if not self._handle_commands(message):
             self._call_listeners('on_plain_message', message)
 
@@ -1273,7 +1462,10 @@ class OrigamiBot:
     def _call_listeners(self, event, *args, **kwargs):
         for listener in self.listeners:
             method = getattr(listener, event)
-            method(*args, **kwargs)
+            if not using_greenlets:
+                method(*args, **kwargs)
+            else:
+                gevent.spawn(method, *args, **kwargs)
 
     def _webhook_loop(self):
         class HandleUpdates(BaseHTTPRequestHandler):
@@ -1288,7 +1480,7 @@ class OrigamiBot:
                 self._set_headers()
                 content_len = int(self.headers.get('content-length', 0))
                 post_body = self.rfile.read(content_len).decode()
-                update = Update(**json.loads(post_body))
+                update = Update.from_json(post_body)
                 HandleUpdates.bot.process_update(update)
 
         HTTPServer((self.webhook, 443), HandleUpdates).serve_forever()
